@@ -4,7 +4,7 @@
 library(nws)
 source('matAdd.R')
 
-# create a sleigh object with default options
+# change launch if you add nodeList parameter
 s = sleigh()
 
 rows = 10
@@ -22,7 +22,7 @@ for (i in 1:10) {
 
     # check the results
     D = A + B
-    if (all.equal(C, D)) {
+    if (isTRUE(all.equal(C, D))) {
         cat('Success\n')
     } else {
         cat('*** Failure ***\n')
